#include <QtWidgets/QApplication>
#include <stdio.h>
#include <QDir>
#include <QDebug>
#include <QString>
#include <QSettings>
#include <QSignalMapper>
#include <QTcpSocket>
#include <QFileInfo>
#include <QJsonDocument>
#include <QStandardPaths>

#ifndef Q_OS_ANDROID
#include <QtSerialPort/QtSerialPort>
#include <QtSerialPort/QSerialPortInfo>
#include <QCommandLineParser>
#endif

#include "actors/alarmtime.h"
#include "microcontroller.h"
#include "ui/mainwindow.h"
#include "sensors/speakersensor.h"
#include "sensors/sunsensor.h"
#include "sensors/ocupancysensor.h"
#include "alarmactions.h"
#include "sensors/sensor.h"
#include "items/itemstore.h"
#include "items/auxitem.h"
#include "items/rgbitem.h"
#include "items/poweritem.h"


#define BAUD  QSerialPort::Baud38400

QJsonDocument* createJsonDocument(const bool pathOption, const QString& filePath)
{
    QFile file;

    #ifndef Q_OS_ANDROID
    if(pathOption && filePath.size() > 0) file.setFileName(filePath);
    else
    #endif
    {
       file.setFileName(QStandardPaths::writableLocation(QStandardPaths::ConfigLocation) + "/shinterface.json");
    }

    file.open(QIODevice::ReadOnly);
    if(!file.isOpen()) std::cerr<<"Can not open config file: "<<filePath.toLatin1().data()<<std::endl;
    else
    {
        QJsonDocument* document = new QJsonDocument(QJsonDocument::fromJson(file.readAll()));
        file.close();
        return  document;
    }
    return new QJsonDocument;
}

void saveJsonDocument(const QJsonDocument * const document, const bool pathOption, const QString& filePath)
{
    QFile file;

    #ifndef Q_OS_ANDROID
    if(pathOption && filePath.size() > 0) file.setFileName(filePath);
    else
    #endif
    {
       file.setFileName(QStandardPaths::writableLocation(QStandardPaths::ConfigLocation) + "/shinterface.json");
    }
    std::cout<<"config file: "<<file.fileName().toLatin1().data()<<std::endl;
    file.open(QIODevice::WriteOnly);
    if(!file.isOpen()) std::cerr<<"Can not open config file: "<<filePath.toLatin1().data()<<std::endl;
    else
    {
        file.write(document->toJson());
        file.close();
    }
}

QSettings* createQSettings(const bool pathOption, const QString& filePath )
{
    QSettings* settings;
    #ifndef Q_OS_ANDROID
    if(pathOption && filePath.size() > 0)
    {
        settings = new QSettings(filePath, QSettings::IniFormat);
        std::cout<<"Config file: "<<settings->fileName().toLatin1().data()<<std::endl;
    }
    else if(pathOption)
    {
        std::cerr<<"Can not open config file: "<<filePath.toLatin1().data()<<std::endl;
        settings = new QSettings;
    }
    else
    #endif
    {
        settings = new QSettings;
    }
    return  settings;
}

void loadItemState(const QJsonDocument * const jsonDocument, QSettings * const settings, Microcontroller* micro, ItemStore* items, PowerItem* power, RgbItem* rgb, AuxItem* aux )
{
    if(jsonDocument)
    {
        const QJsonObject jsonObject = jsonDocument->object();
        items->load(jsonObject, micro);
        power->load(jsonObject["Power"].toObject());
        if(jsonObject["Items"].toArray().size() >= 2)
        {
            rgb->load(jsonObject["Items"].toArray()[0].toObject());
            aux->load(jsonObject["Items"].toArray()[1].toObject());
        }
    }
    else
    {
        items->load("Items", settings, micro);
        power->load("Power", settings);
    }
}

int main(int argc, char *argv[])
{
    QApplication a(argc, argv);

    //set info
    QCoreApplication::setOrganizationName("UVOS");
    QCoreApplication::setOrganizationDomain("uvos.xyz");
    QCoreApplication::setApplicationName("SHinterface");
    QCoreApplication::setApplicationVersion("0.6");

    QDir::setCurrent(a.applicationDirPath());

    //parse comand line
    #ifndef Q_OS_ANDROID
    QCommandLineParser parser;
    parser.setApplicationDescription("Smart Home Interface");
    parser.addHelpOption();
    parser.addVersionOption();
    QCommandLineOption tcpOption(QStringList() << "t" << "tcp", QCoreApplication::translate("main", "Use Tcp connection"));
    parser.addOption(tcpOption);
    QCommandLineOption hostOption(QStringList() << "H" << "host", QCoreApplication::translate("main", "Set server host ip addres"), "adress");
    parser.addOption(hostOption);
    QCommandLineOption portOption(QStringList() << "p" << "port", QCoreApplication::translate("main", "Set server Port in TCP mode or Serial port in serial mode"), "port");
    parser.addOption(portOption);
    QCommandLineOption serialOption(QStringList() << "s" << "serial", QCoreApplication::translate("main", "Use serial connection"));
    parser.addOption(serialOption);
    QCommandLineOption baudOption(QStringList() << "b" << "baud", QCoreApplication::translate("main", "Set Baud Rate"));
    parser.addOption(baudOption);
    QCommandLineOption settingsPathOption(QStringList() << "c" << "config", QCoreApplication::translate("main", "Set config file"), "configFilePath");
    parser.addOption(settingsPathOption);
    QCommandLineOption secondaryOption(QStringList() << "e" << "secondary", QCoreApplication::translate("main", "Set if instance is not main instance"));
    parser.addOption(secondaryOption);
    QCommandLineOption qsettingsOption(QStringList() << "q" << "qsettigns", QCoreApplication::translate("main", "Set if jsettings file format should be used."));
    parser.addOption(secondaryOption);
    parser.process(a);
    #endif

    QSettings* settings = nullptr;
    QJsonDocument* jsonDocument = nullptr;
    if(parser.isSet(qsettingsOption)) settings = createQSettings(parser.isSet(settingsPathOption), parser.value(settingsPathOption));
    else jsonDocument = createJsonDocument(parser.isSet(settingsPathOption), parser.value(settingsPathOption));


    //connect to microcontoler
    Microcontroller micro;

    #ifndef Q_OS_ANDROID
    if(parser.isSet(tcpOption))
    {
        QTcpSocket* microSocket = new QTcpSocket;

        int port = 6856;
        if(parser.isSet(portOption)) port =  parser.value(portOption).toInt();

        QString host("127.0.0.1");
        if(parser.isSet(hostOption)) host = parser.value(hostOption);
        std::cout<<"connecting to "<<host.toStdString()<<':'<<port<<'\n';
        microSocket->connectToHost(host, port, QIODevice::ReadWrite);
        if(!microSocket->waitForConnected(1000))
        {
            std::cout<<"Can not connect to to Server.\n";
            return 1;
        }
        micro.setIODevice(microSocket);
    }
    else
    {
        QSerialPort* microPort = new QSerialPort;
        if(parser.isSet(portOption)) microPort->setPortName(parser.value(portOption));
        else microPort->setPortName("ttyUSB0");

        if(parser.isSet(portOption)) microPort->setBaudRate(parser.value(baudOption).toInt());
        else microPort->setBaudRate(BAUD);

        if(!microPort->open(QIODevice::ReadWrite)) std::cout<<"Can not open serial port "<<microPort->portName().toStdString()<<". Continueing in demo mode"<<'\n';
        else micro.setIODevice(microPort);
    }
    #else
        QTcpSocket* microSocket = new QTcpSocket;
        microSocket->connectToHost("10.0.0.1", 6856, QIODevice::ReadWrite);
        if(!microSocket->waitForConnected(1000))
        {
            std::cout<<"Can not connect to to Server.\n";
            return 1;
        }
        micro.setIODevice(microSocket);
    #endif


    AlarmActions alarmActions(&a, &micro);

    //sensors
    SpeakerSensorSource livingroomSpeakerSensorSource("Livingroom Speakers");
    SunSensorSource     sunSensorSource(49.884450, 8.650536);
    OcupancySensorSource ocupancySensor;

    SensorStore sensors;
    QObject::connect(&micro, &Microcontroller::gotSensorState, &sensors, &SensorStore::sensorGotState);
    QObject::connect(&livingroomSpeakerSensorSource, &SpeakerSensorSource::stateChanged, &sensors, &SensorStore::sensorGotState);
    QObject::connect(&sunSensorSource, &SunSensorSource::stateChanged, &sensors, &SensorStore::sensorGotState);
    QObject::connect(&sensors, &SensorStore::sensorChangedState, &ocupancySensor, &OcupancySensorSource::sensorEvent);
    QObject::connect(&ocupancySensor, &OcupancySensorSource::stateChanged, &sensors, &SensorStore::sensorGotState);

    QMetaObject::invokeMethod(&livingroomSpeakerSensorSource, "run", Qt::QueuedConnection);
    sunSensorSource.run();

    //item store
    ItemStore items(&sensors);
    QObject::connect(&micro, &Microcontroller::gotRelayList, &items, &ItemStore::addItems);
    QObject::connect(&micro, &Microcontroller::itemChanged, &items, &ItemStore::itemStateChanged);

    //Power Item
    PowerItem powerItem(&sensors, &a, 5487423, "Power");
    QObject::connect(&powerItem, &PowerItem::stateChanged, &sensors, &SensorStore::sensorGotState);
    powerItem.emmitSensor();

    //mainwindow
    MainWindow w(&micro, &powerItem, &items, &sensors);
    QObject::connect(&micro, SIGNAL(textRecived(QString)), &w, SLOT(changeHeaderLableText(QString)));
    if(!micro.connected()) w.changeHeaderLableText("No io debug only!");

    //special items
    std::shared_ptr<RgbItem> rgbItem(new RgbItem(&sensors, &micro, 5487422, "Rgb Lights"));
    items.addItem(rgbItem);

    std::shared_ptr<AuxItem> auxItem(new AuxItem(&sensors, &micro, 5487421, "Desk Light"));
    items.addItem(auxItem);

    QMetaObject::invokeMethod(&micro, "run", Qt::QueuedConnection  );

    loadItemState(jsonDocument, settings, &micro, &items, &powerItem, rgbItem.get(), auxItem.get());

    #ifndef Q_OS_ANDROID
    if(!parser.isSet(secondaryOption))
    {
        Item::secondaryFlag = false;
    }
    else
    {
        Item::secondaryFlag = true;
    }
    #endif

    w.show();

    //micro.requestState();
    int retVal = a.exec();

    if(settings)
    {
        items.store("Items", settings);
        powerItem.store("Power/", settings);
        settings->sync();
    }
    else
    {
        QJsonObject itemsJsonObject;
        items.store(itemsJsonObject);
        QJsonObject powerObject;
        powerItem.store(powerObject);
        itemsJsonObject.insert("Power", powerObject);
        jsonDocument->setObject(itemsJsonObject);
        saveJsonDocument(jsonDocument, parser.isSet(settingsPathOption), parser.value(settingsPathOption));
    }
    if(settings) delete settings;
    else delete jsonDocument;
    return retVal;
}

